package calc;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * Constraint represent a Constraint, either hard coded or derived from a configuration
 * @author sjoerd
 */
public class Constraint implements Serializable {
    private ArrayList<Term> terms;
    
    public Constraint(){
        terms = new ArrayList<Term>();
    }
    
    public void addTerm(int constant,int coefficient){
        terms.add(new Term(constant, coefficient));
    }
            
    public boolean check(double x, double alpha){
        return (sum(x,alpha) <= 1.0);
    }
    
    public double sum(double x, double alpha){
        double value = 0;
        for(Term t:terms){
            value += t.calc(x, alpha);
        }
        return value;
    }
    
    /**
     * calculate the branchingvector in human readable format
     * @return 
     */
    public String getVector(){
        String res="";
        Iterator<Term> it = terms.iterator();
        while(it.hasNext()){
            res+=it.next().print();
            if(it.hasNext()){
                res+=", ";
            }
        }
        return "("+res+")";
    }
    
    /**
     * The constraint can be printed in mathematica syntax. Tested with mathematica 8.0
     * Multiple constraints can be plotted using:
     * xlo = 1.5;
     * xhi = 1.7;
     * alo = 0.4;
     * ahi = 0.7;
     * opa = 0.000;
     * Show[<constraints>,FrameLabel -> {"x", "\[Alpha]"},PlotRange -> {{xlo, xhi}, {alo, ahi}},ImageSize -> 600,BaseStyle -> {FontFamily -> "Helvetica"}  ]   
     * @return a text representation that can be copy-pasted into matthematica
     */
    public String mathematicaPrint(){
        String res="RegionPlot[ ";
                //x^-(4 + 2 \\[Alpha]) + x^-(1 - \\[Alpha]) ";
        Iterator<Term> it = terms.iterator();
        while(it.hasNext()){
            res += "x^-(";
            //Term t = it.next();
            res+=it.next().print("\\[Alpha]");
            res +=")";
            if(it.hasNext()){
                res+=" + ";
            }
        }
        return res+" < 1, {x, 1, 2}, {\\[Alpha], 0,   1}, PlotStyle -> Opacity[0], BoundaryStyle -> Directive[Black]],";
    }
}

/**
 * Inner class Term represents one term in a constraint like 2+α
 * @author Sjoerd Timmer
 */
class Term implements Serializable {
    int constant;
    int coefficient;
    
    public Term(int constant,int coefficient){
        this.constant = constant;
        this.coefficient = coefficient;
    }
    
    /**
     * Print the term in human readable format
     * @param varname how to print α
     * @return the string representation
     */
    public String print(String varname){
        String res = "";
        res+=constant;
        if(coefficient!=0)
            res+=(coefficient>0?"+":"-")+(Math.abs(coefficient)==1?"":Math.abs(coefficient))+varname;
        return res;
    }
    
    public String print(){
        return print("α");
    }
    
    public double calc(double x,double alpha){
        double exp = (double)constant + (double)coefficient*alpha;
        return Math.pow(x, -exp);
    }
}