package kayles;

import calc.Constraint;
import calc.Optimizer;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import javax.swing.JTree;
import javax.swing.tree.TreeNode;
import utils.DualHashBidiMap;
import utils.Range;
import utils.Tuple;


/**
 * Representation of one configuration. I.e. a set of nodes and a numbering on those nodes
 * @author sjoerd
 */
public class Configuration implements TreeNode,Serializable {
    public ArrayList<Node> nodes;
    public DualHashBidiMap<Integer,Node> ids;
    private Constraint constr;
    public Configuration iso;
    transient JTree tree;
    public double expandability = 1;
    
    // maintain tree-structure, only for GUI
    private Configuration parent;
    private ArrayList<Configuration> children = new ArrayList<Configuration>();
    private HashSet<Node> expanded = new HashSet<Node>();
    private String desc = "";
    
    public static final long serialVersionUID  = 7350453611156259534L;
    
    /**
     * Construct a new configuration with the given description
     * @param name how to name this in the GUI
     */
    public Configuration(String name){
        nodes = new ArrayList<Node>();
        ids = new DualHashBidiMap<Integer, Node>();
        desc = name;
    }
    
    /**
     * Construct a new configuration with the given description. The expandability symbolises the fact that we want to hint the program that it is not wise to expand this configuration until the solution is below the given value
     * @param name description for the GUI
     * @param expandability hint, as to whether or not to expand this node.
     */
    public Configuration(String name,double expandability){
        this(name);
        this.expandability = expandability;
    }
    
    /**
     * getter for the array of children
     * @return the array of children
     */
    public ArrayList<Configuration> getChildren(){
        return children;
    }
    
    /**
     * Transforms the branchingvector for this case into a Constraint
     * @see Constraint
     * @return the Constraint that follows from this case
     */
    public Constraint getConstraint(){
        if(constr==null){
            constr=new Constraint();
            for(Tuple red:getBranchingVector()){
            //    System.out.println(""+red.num+(red.alpha>0?"+":"-")+Math.abs(red.alpha)+"*alpha");
                constr.addTerm(red.num, red.alpha);
            }
        }
        return constr;
    }
    
    /**
     * Getter for the node list
     * @return the node list in this configuration
     */
    public ArrayList<Node> getNodes(){
        return nodes;
    }
    
    /**
     * adds a node with the specified properties to the graph
     * @param type RED,GREEN,WHITE or BLACK
     * @param lo_red minimum number of external red neighbors
     * @param hi_red maximum number of external red neighbors
     * @param lo_white minimum number of external white neighbors
     * @param hi_white maximum number of external white neighbors
     * @return the auto generated number for the new node
     */
    public int addNode(NodeType type,int lo_red,int hi_red, int lo_white, int hi_white){
        int id = 0;
        while(true){
            if(!ids.isSet(++id)) break; 
        }
        addNode(id,type,lo_red,hi_red,lo_white,hi_white);
        return id;
    }
    
    /**
     * adds a node with the specified properties to the graph
     * @param id the desired id, if this is not available, results are unpredictable!
     * @param type RED,GREEN,WHITE or BLACK
     * @param lo_red minimum number of external red neighbors
     * @param hi_red maximum number of external red neighbors
     * @param lo_white minimum number of external white neighbors
     * @param hi_white maximum number of external white neighbors
     */
    public void addNode(int id, NodeType type,int lo_red,int hi_red, int lo_white, int hi_white){
        Node n = new Node(id,type);
        n.setEnv(lo_red, hi_red, lo_white, hi_white);
        nodes.add(n);
        ids.put(id, n);
    }
    
    /**
     * Connect the nodes with the given numbers
     * @param id1 id of the first node
     * @param id2 id of the second node
     */
    public void connect(int id1,int id2){
        Node n = ids.getValue(id1);
        Node m = ids.getValue(id2);
        n.add(m);
        m.add(n);
    }
    
    /**
     * Create an exact replication of this configuration, all nodes are duplicated as well and connected in the same way
     * @param newname the desired new name for the new configuration
     * @return the new configuration
     */
    public Configuration duplicate(String newname){
        Configuration newconf = new Configuration(newname,expandability);
        // clone the nodes
        for(Node n:nodes){
            newconf.addNode(ids.getKey(n), n.type, n.ext_red.lo, n.ext_red.hi, n.ext_white.lo, n.ext_white.hi);
        }
        for(Node n:nodes){
            // get the equivalent node 
            Node m = newconf.ids.getValue(ids.getKey(n));
            // for all neighbors of n
            for(Node nbn:n.getNeighbors()){
                // get the equivalent of the neighbor
                Node nbm = newconf.ids.getValue(ids.getKey(nbn));
                m.add(nbm);
            }
        }
        return newconf;
    }
    
    /**
     * Calculates the set of possible reductions
     * @return a HashSet of Reductions
     */
    public HashSet<Reduction> getBranchings(){
        // determine inner nodes
        // as arralist since it will mostly be iterated
        ArrayList<Node> innernodes = new ArrayList<Node>();
        for(Node n:nodes){
            if(n.ext_white.hi==0&&n.type==NodeType.WHITE){ // TODO: moet ext_red ook 0 zijn??
                innernodes.add(n);
            }
        }
        
        ArrayList<HashSet<Node>> iss = getAllIdependentSets(innernodes);
        
        // generate a set of possible reductions
        HashSet<Reduction> reductions = new HashSet<Reduction>();
        for(HashSet<Node> is:iss){
            Reduction r = getReduction(innernodes, is);
            reductions.add(r);
        }
        return reductions;
    }
    
    /**
     * Calculate the branching vector, this uses the reductions returned by getBranchings()
     * @return the branching vector
     */
    public ArrayList<Tuple> getBranchingVector(){
        HashSet<Reduction> reductions = getBranchings();
        // for each unique reduction calculate the gain
        ArrayList<Tuple> res = new ArrayList<Tuple>();
        for(Reduction reduction:reductions){
            res.add(reduction.calcGain());
        }
        return res;
    }
    
    /**
     * Given the set of inner nodes and the set that is to be selected, calculate the resulting coloring of the graph.
     * This method removes isolated nodes and applies "rule 1" automatically
     * @param inner set of inner nodes
     * @param selection set of nodes that is to be selected
     * @return a reduction(i.e. map from node to coloring)
     */
    public Reduction getReduction(ArrayList<Node> inner,HashSet<Node> selection){
        // keep a virtual assignment of colors
        Reduction reduction = new Reduction();
        for(Node n:nodes){
            reduction.put(n, n.type);
        }
        
        // change the colors of the innernodes
        for(Node n:inner){
            if(selection.contains(n)){ // only if it was not yet deleted
                reduction.put(n, NodeType.BLACK);
                for(Node m:n.getNeighbors()){
                    reduction.put(m, NodeType.BLACK);
                }
            }else{
                if(reduction.get(n)==NodeType.WHITE){
                    reduction.put(n, NodeType.RED);
                }
            }
        }
        // if an innernode is no longer connected to the borders it gets black as well
        HashSet<Node> reachable = new HashSet<Node>();
        // gets the non-deleted non-inner nodes as initial content
        for(Node n:nodes){
            if(reduction.get(n)!=NodeType.BLACK && (n.ext_white.hi!=0 || n.ext_red.hi!=0) ) reachable.add(n);
        }
        boolean done =false;
        HashSet newreachables = new HashSet();
        while(!done){
            newreachables.clear();
            for(Node n:reachable){
                for(Node m:n.getNeighbors()){
                    if(reduction.get(m) !=NodeType.BLACK){
                        newreachables.add(m);
                    }
                }
            }
            done = !reachable.addAll(newreachables);
        }
        // now color the unreachable nodes black
        for(Node n:inner){
            if(!reachable.contains(n)){
                reduction.put(n, NodeType.BLACK);
            }
        }
        
        // apply rule 1
        for(Node n:nodes){
            applyRule1(n, reduction);
        }
        // TODO: the other rules?
        
        return reduction;
    }
    
    
    /**
     * This method returns all possible independent sets.
     * BEWARE: this method breaks the argument apart and rebuilds it, do not use it (for instance in another thread) meanwhile!!
     * @param the set of which we want the independent sets to be a subset, typically the selection
     * @return a list of sets of nodes that can be selected simultaneously
     */
    private ArrayList<HashSet<Node>> getAllIdependentSets(ArrayList<Node> subset){
        ArrayList<HashSet<Node>> res = new ArrayList<HashSet<Node>>();
        if(subset.isEmpty()){
            res.add(new HashSet<Node>());
        }else{
            Node n = subset.get(0);
            ArrayList<Node> neighbors = new ArrayList<Node>(n.getNeighbors());
            neighbors.retainAll(subset);
            subset.remove(n);
            res = getAllIdependentSets(subset);
            if(n.type==NodeType.WHITE){
                subset.removeAll(neighbors);
                ArrayList<HashSet<Node>> withn = getAllIdependentSets(subset);
                for(HashSet<Node> al:withn){
                    al.add(n);
                }
                res.addAll(withn);
                subset.addAll(neighbors);
            }
            subset.add(n);
        }
        return res;
    } 
    
    /**
     * Tries to apply "rule 1" to the given node in the given reduction
     * @param n the node to try to apply rule 1 to
     * @param r the reduction so far (used to determine virtual colors)
     */
    private void applyRule1(Node n, Reduction r){
        if(r.get(n)==NodeType.RED && n.ext_white.hi==0){
            boolean ruleapplies = true;
            for(Node m:n.getNeighbors()){
                if(r.get(m)==NodeType.WHITE){
                    ruleapplies=false;
                }
            }
            if(ruleapplies){ 
                r.put(n, NodeType.GREEN);
                for(Node m:n.getNeighbors()){
                    applyRule1(m, r);
                }
            }
        }
    }

     /**
     * Check if the configuration has a connection to the outside world
     * @return true if no connection exists
     */
    public boolean isClosed(){
        for(Node n:nodes){
            if(n.ext_white.hi>0){
                return false;
            }
        }
        return true;
    }
    
    /**
     * Expand one node, generating all possible configurations for it's white neighbors
     * @param id the id of the node to expand
     * @param numexpanded the maximum number of neighbors to generate
     * @return a list of newly created configurations
     */
    public ArrayList<Configuration> expandOnWhite(int id, int numexpanded) {
        // don't expand isomorphic nodes
        if(iso!=null){
            return new ArrayList<Configuration>();
        }
        
        //System.out.println("expanding!");
        ArrayList<Configuration> list = new ArrayList<Configuration>();
        Node n = ids.getValue(id);
        if(n==null){
            return new ArrayList<Configuration>();
        }
        
        // generate the list of current neighbors that is acceptable as a new neighbor
        ArrayList<Integer> availableneighbors = new ArrayList<Integer>(ids.getKeySet());
        // remove non-white nodes from that list
        for(Node nb:nodes){
            // remove:
            //    non-white
            //    self
            //    saturated nodes
            if(nb.type!=NodeType.WHITE || ids.getValue(id)==nb || nb.ext_white.hi==0){
                availableneighbors.remove(ids.getKey(nb));
            }
        }
        // remove current neighbors from that list
        for(Node nb:n.getNeighbors()){
            availableneighbors.remove(ids.getKey(nb));
        }
        
        // no need to expand more than ext_white.hi
        numexpanded = Math.min(numexpanded, n.ext_white.hi);
        // iterate over all subsets of that and all possible numbers of allowed white neighbors
        for(int wn=n.ext_white.lo;wn<=numexpanded;wn++){
            for(ArrayList<Integer> sub:getAllSubsets(availableneighbors)){
                // build the config where n gets connected to the subset
                // and some new white nodes if neccessary to get to a total of wn
                if(sub.size()>wn){
                    // nothing to do here
                }else{
                    // create a new configuration
                    String name = "expand "+id+": ";
                    for(Integer i:sub){
                        name+=""+i+", ";
                    }
                    name+="+"+(wn-sub.size())+"new";
                    Configuration nc = this.duplicate(name);
                    
                    // connect newn to the subset
                    for(Integer i:sub){
                        nc.connect(i, id);
                        Node other = nc.ids.getValue(i);
                        other.ext_white.add(-1); // that node now has one less
                    }
                    // connect the node to some new nodes
                    for(int i=sub.size();i<wn;i++){
                        int newid = nc.addNode(NodeType.WHITE, 0, Range.INF, 0, 3);
                        nc.connect(id, newid);
                    }
                    
                    // cleaning up:
                    Node newn = nc.ids.getValue(id);
                    
                    if(wn==availableneighbors.size()+numexpanded){
                        // this is the wn-or-more phase
                        newn.ext_white.add(-wn);
                        // nodes are always expanded to have two new white neighbors
                        // and for white nodes this is the maximum
                        newn.ext_white.hi=newn.type==NodeType.WHITE?0:newn.ext_white.hi;
                    }else{
                        newn.ext_white.hi=0;
                    }
                    newn.ext_white.lo=0;
                    
                    // make it a tree element
                    children.add(nc);
                    nc.parent=this;
                    nc.setTree(tree);
                    
                    // add it to the return-list
                    list.add(nc);
                }
            }
        }   
        return list;
    }
    
    /**
     * 
     * @param id node to expand
     * @param desirednewnodes maximum number of new neighbors to create
     * @return a list of newly created configurations
     */
    public ArrayList<Configuration> expandOnRed(int id, int desirednewnodes) {
        // don't expand isomorphic nodes
        expanded.add(ids.getValue(id));
        
        if(iso!=null){
            return new ArrayList<Configuration>();
        }
        
        //System.out.println("expanding!");
        ArrayList<Configuration> list = new ArrayList<Configuration>();
        Node n = ids.getValue(id);
        if(n==null){
            return new ArrayList<Configuration>();
        }
        
        // generate the list of current neighbors that is acceptable as a new neighbor
        ArrayList<Integer> availableneighbors = new ArrayList<Integer>(ids.getKeySet());
        // remove non-red nodes from that list
        for(Node nb:nodes){
            // remove:
            //    non-red
            //    self
            //    saturated nodes
            if(nb.type!=NodeType.RED || ids.getValue(id)==nb || (n.type==NodeType.RED && nb.ext_red.hi==0) || (n.type==NodeType.WHITE && nb.ext_white.hi==0)){
                availableneighbors.remove(ids.getKey(nb));
            }
        }
        // remove current neighbors from that list
        for(Node nb:n.getNeighbors()){
            availableneighbors.remove(ids.getKey(nb));
        }
        
        // no need to expand more than ext_white.hi
        desirednewnodes = Math.min(desirednewnodes, n.ext_red.hi);
        // iterate over all subsets of that and all possible numbers of allowed red neighbors
        // to achieve the desired number of new nodes, we need desirednewnodes+#availableneighbors
        for(int rn=n.ext_red.lo;rn<=availableneighbors.size()+desirednewnodes;rn++){
            for(ArrayList<Integer> sub:getAllSubsets(availableneighbors)){
                // build the config where n gets connected to the subset
                // and some new red nodes if neccessary to get to a total of rn
                if(sub.size()>rn){
                    // nothing to do here
                }else{
                    // create a new configuration
                    String name = "expand-reds "+id+": ";
                    for(Integer i:sub){
                        name+=""+i+", ";
                    }
                    name+="+"+(rn-sub.size())+"new";
                    Configuration nc = this.duplicate(name);
                    
                    // connect newn to the subset
                    for(Integer i:sub){
                        nc.connect(i, id);
                        Node other = nc.ids.getValue(i);
                        // that node now has one less
                        if(n.type==NodeType.WHITE){
                            other.ext_white.add(-1);
                        }else{
                            other.ext_red.add(-1);
                        }
                    }
                    // connect the node to some new, red nodes
                    for(int i=sub.size();i<rn;i++){
                        int newid = nc.addNode(NodeType.RED, 0, Range.INF, 0, 3);
                        nc.connect(id, newid);
                    }
                    
                    // cleaning up:
                    Node newn = nc.ids.getValue(id);
                    newn.ext_red.hi= (rn==availableneighbors.size()+desirednewnodes)?Range.INF:0;
                    newn.ext_red.lo=0;
                    
                    // make it a tree element
                    children.add(nc);
                    nc.parent=this;
                    nc.setTree(tree);
                    
                    // add it to the return-list
                    list.add(nc);
                }
            }
        }   
        return list;
    }
    
    /**
     * Apply the heuristic to the given configuration
     * @return true if any of the cases could be applied, false otherwise
     */
    public boolean heuristicExpand(){
        for(int id:ids.getKeySet()){
            if(ids.getValue(id).type==NodeType.WHITE && !expanded.contains(ids.getValue(id))){
                if(ids.getValue(id).ext_white.hi==0 && ids.getValue(id).ext_red.hi>0){
                    Optimizer.staticInstance.allconfigs.addAll(expandOnRed(id, 1));
                    return true;
                }
            }
        }
        for(int id:ids.getKeySet()){
            if(ids.getValue(id).type==NodeType.WHITE){
                if(ids.getValue(id).ext_white.hi>0){
                    Optimizer.staticInstance.allconfigs.addAll(expandOnWhite(id, 2));
                    return true;
                }
            }
        }
        for(int id:ids.getKeySet()){
            if(ids.getValue(id).type==NodeType.RED){
                if(ids.getValue(id).ext_white.hi>0){
                    Optimizer.staticInstance.allconfigs.addAll(expandOnWhite(id, 2));
                    return true;
                }
            }
        }
        System.err.println("No expansion could be make!!!");
        return false;
    }
    
    
    /**
     * Recursively expand this configuration until it has no tight children
     */
    public void recursiveCleverExpand(){
        if(!this.hasTightChildren()){
            return ;
        }
    //    boolean res = true;
        if(this.isLeaf()){
            if(!heuristicExpand()){
                return ;
            }
            Optimizer.staticInstance.allconfigs.remove(this);
           
        }else{
            for(Configuration child:getChildren()){
                child.recursiveCleverExpand();
            }
        }
        return ;
    }
    
    /**
     * Check for an isomorphism
     * @param other the other configuration to check against
     * @return true if this and other are isomorphic
     */
    public boolean findIsomorphism(Configuration other){
        
        if(this.nodes.size()!=other.nodes.size()){
            return false;
        }
      //  System.out.println("isomorphism check: "+this.toString()+" <-> "+other.toString());
      //  System.out.println("checking node-count: OK");
        
        // create a mapping from group to sets of nodes
        HashMap<Integer,HashSet<Node>> groups_this = new HashMap<Integer, HashSet<Node>>();
        HashMap<Integer,HashSet<Node>> groups_other = new HashMap<Integer, HashSet<Node>>();
        
        
        for(Node n:this.nodes){
            int cat = n.getGroup();
            if(!groups_this.containsKey(cat)){
                groups_this.put(cat, new HashSet<Node>());
            }
            HashSet bin = groups_this.get(cat);
            bin.add(n);
        }
        // also for the other graph
        for(Node n:other.nodes){
            int cat = n.getGroup();
            if(!groups_other.containsKey(cat)){
                groups_other.put(cat, new HashSet<Node>());
            }
            HashSet bin = groups_other.get(cat);
            bin.add(n);
        }
        // check that the same groups were created
        Set<Integer> thesekeys = groups_this.keySet();
        Set<Integer> thosekeys = groups_other.keySet();
        if(!thesekeys.containsAll(thosekeys)||!thosekeys.containsAll(thesekeys)){
            return false;
        }
        // check that the groups have equal size
        for(Integer i:groups_this.keySet()){
            HashSet<Node> thesenodes = groups_this.get(i);
            HashSet<Node> thosenodes = groups_other.get(i);
            if(thesenodes.size()!=thosenodes.size()){
                return false;
            }
        }
     //   System.out.println("checking subgroups: OK");
        
        // generate the possible isomorphisms for each group
        HashMap<Integer,HashSet<HashMap<Node,Node>>> isospergroup = new HashMap<Integer, HashSet<HashMap<Node, Node>>>();
        for(Integer i:groups_this.keySet()){
            // make all possible permutations
            HashSet<HashMap<Node,Node>> isosforthisgroup = new HashSet<HashMap<Node, Node>>();
            for(ArrayList<Node> perm:getPermutations(groups_other.get(i))){
                HashMap<Node,Node> newiso = new HashMap<Node, Node>();
                Iterator<Node> these = groups_this.get(i).iterator();
                for(Node nd:perm){
                    newiso.put(these.next(), nd);
                }
                // only store this iso if it is consistent in itself
                if(checkPartialIsomorphism(newiso)){
                    isosforthisgroup.add(newiso);
                }
            }
            isospergroup.put(i, isosforthisgroup);
        }
        
        // now check all combinations of iso's from all groups
        for(HashMap<Node,Node> theiso:combineIsomorphisms(isospergroup)){
            if(checkIsomorphism(theiso)){
                this.iso = other;
                return true;
            }
        }
        
        return false;
    }
    
    /**
     * combines the isomorphisms for a couple of groups to a set of all combinations that covers all possible isomorphisms on the complete graphs
     * BEWARE: this method breaks the argument apart and rebuilds it, do not use it (for instance in another thread) meanwhile!!
     * @param isospergroup the groups, each as a Set of partial isomorphism
     * @return a set of complete isomorphisms
     */
    public HashSet<HashMap<Node,Node>> combineIsomorphisms(HashMap<Integer,HashSet<HashMap<Node,Node>>> isospergroup){
        /* TODO: can performance increase by changing multiplication order? */
        HashSet<HashMap<Node,Node>> res = new HashSet<HashMap<Node, Node>>();
        if(isospergroup.size()==1){
            res.addAll(isospergroup.get(isospergroup.keySet().iterator().next()));
            return res;
        }

        // temporarily remove the first group
        int firstcat = isospergroup.keySet().iterator().next();
        HashSet<HashMap<Node,Node>> firstset = isospergroup.remove(firstcat);
        HashSet<HashMap<Node,Node>> restresult = combineIsomorphisms(isospergroup);
        for(HashMap<Node,Node> iso1:firstset){
            for(HashMap<Node,Node> iso2:restresult){
                HashMap<Node,Node> newiso = new HashMap<Node, Node>();
                newiso.putAll(iso1);
                newiso.putAll(iso2);
                res.add(newiso);
            }
        }

        // add the first group back on
        isospergroup.put(firstcat, firstset);
        return res;
    }
    
    /**
     * Get a list of permutations. Used by the isomorphism checker
     * @param ns list of nodes
     * @return all possible orderings of the nodes in ns
     */
    public ArrayList<ArrayList<Node>> getPermutations(HashSet<Node> ns){
        if(ns.size()<=1){
            ArrayList<ArrayList<Node>> res = new ArrayList<ArrayList<Node>>();
            res.add(new ArrayList<Node>(ns));
            return res;
        }
        ArrayList<ArrayList<Node>> res = new ArrayList<ArrayList<Node>>();
        for(Node last:ns){
            HashSet<Node> smaller = (HashSet<Node>)ns.clone();
            smaller.remove(last);
            
            for(ArrayList<Node> others:getPermutations(smaller)){
                ArrayList<Node> newperm = new ArrayList<Node>(others);
                newperm.add(last);
                res.add(newperm);
            }
        }
        return res;
    }
    
    /**
     * Checks whether a mapping from nodes to nodes could be part of an isomorphism
     * @param iso the mapping from node to node
     * @return true if this is consistent, false otherwise
     */
    public boolean checkPartialIsomorphism(HashMap<Node,Node> iso){
        //for all edges
        for(Node n:iso.keySet()){
            for(Node m:n.getNeighbors()){
                // check if the mapping of both endpoints are connected (only if m in this group)
                if(iso.keySet().contains(m) && !iso.get(n).getNeighbors().contains(iso.get(m))){
                    return false;
                }
            }
        }
        return true;
    }
    
    
    /**
     * Checks a complete isomorphism
     * @param iso the mapping to test
     * @return true, if and only if the isomorphism is correct
     */
    public boolean checkIsomorphism(HashMap<Node,Node> iso){
      //for all edges
        for(Node n:iso.keySet()){
            for(Node m:n.getNeighbors()){
                // check if the mapping of both endpoints are connected
                if(!iso.get(n).getNeighbors().contains(iso.get(m))){
                    return false;
                }
            }
        }
        return true;
    }
    
    /**
     * set the tree of which this configuration is part. Used for the GUI
     * @param tree the tree
     */
    public void setTree(JTree tree){
        this.tree = tree;
    }
    
    /**
     * returns all subsets of a list
     * BEWARE: this method breaks the argument apart and rebuilds it, do not use it (for instance in another thread) meanwhile!!
     * @param ns the list
     * @return a list of subsets
     */
    public ArrayList<ArrayList<Integer>> getAllSubsets(ArrayList<Integer> ns){
        ArrayList<ArrayList<Integer>> res = new ArrayList<ArrayList<Integer>>();
        if(ns.isEmpty()){
            res.add(new ArrayList<Integer>());
        }else{
            int n = ns.get(0);
            ns.remove((Object)n);
            // calculate all subsets without n
            ArrayList<ArrayList<Integer>> withoutn = getAllSubsets(ns);
            // add all subsets twice to the result, 
            for(ArrayList<Integer> list:withoutn){
                res.add(new ArrayList<Integer>(list));
                list.add(n);
                res.add(list);
            }
            ns.add(n);
        }
        return res;
    }
    

    @Override
    public String toString(){
        return desc;
    }

    /**
     * For the GUI: determine whether this configuration is tight or has tight children
     * @return true if it has, false otherwise
     */
    public boolean hasTightChildren(){
        if(Optimizer.staticInstance.tightConfigs.contains(this)){
            return true;
        }
        for(Configuration child:children){
            if(child.hasTightChildren()){
                return true;
            }
        }
        return false;
    }
    
    /**
     * just for statistical reference
     * @return the depth of the expansion tree below this config
     */
    public int getDepth(){
        int max = 0;
        for(Configuration c:children){
            max = Math.max(max, 1+c.getDepth());
        }
        return max;
    }
    
    /**
     * just for statistical reference
     * @return the number of leaf-configurations below this config
     */
    public int countDescendants(){
        if(isLeaf()){
            return 1;
        }else{
            int res = 0;
            for(Configuration c:children){
                res+= c.countDescendants();
            }
            return res;
        }
    }
    
    /**
     * Print the leaf-configurations below this one
     */
    public void printCases(){
        if(isLeaf()){
            System.out.println(getConstraint().mathematicaPrint());
        }else{
            for(Configuration c:children){
                c.printCases();
            }
        }
    }
    
    // implement the TreeNode methods for thte GUI
    @Override
    public TreeNode getChildAt(int childIndex) {
        return children.get(childIndex);
    }

    @Override
    public int getChildCount() {
        return children.size();
    }

    @Override
    public TreeNode getParent() {
        return parent;
    }

    @Override
    public int getIndex(TreeNode node) {
        return children.indexOf(node);
    }

    @Override
    public boolean getAllowsChildren() {
        return true;
    }

    @Override
    public boolean isLeaf() {
        return children.isEmpty();
    }

    @Override
    public Enumeration children() {
        return Collections.enumeration(children);
    }
}
